"""
'dashboard.py'
=========================================
Creates a dashboard with 3 blocks and feed it data

Author(s): Doug Zobel
"""
from time import sleep
from random import randrange
from Adafruit_IO import Client, Feed, Block, Dashboard, Layout

# Set to your Adafruit IO key.
# Remember, your key is a secret,
# so make sure not to publish it when you publish this code!
ADAFRUIT_IO_USERNAME = ''

# Set to your Adafruit IO username.
# (go to https://accounts.adafruit.com to find your username)
ADAFRUIT_IO_KEY = ''

# Create an instance of the REST client.
aio = Client(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

# Create a new feed named 'Dashboard Data' under the default group
feed = aio.create_feed(Feed(name="Dashboard Data"), "default")

# Fetch group info (group.id needed when adding feeds to blocks)
group = aio.groups("default")

# Create a new dasbhoard named 'Example Dashboard'
dashboard = aio.create_dashboard(Dashboard(name="Example Dashboard"))

# Create a line_chart
linechart = Block(name="Linechart Data",
                  visual_type = 'line_chart',
                  properties = {
                      "gridLines": True,
                      "historyHours": "2"},
                  block_feeds = [{
                      "group_id": group.id,
                      "feed_id":  feed.id
                  }])
linechart = aio.create_block(dashboard.key, linechart)

# Create a gauge
gauge = Block(name="Gauge Data",
              visual_type = 'gauge',
              block_feeds = [{
                  "group_id": group.id,
                  "feed_id":  feed.id
              }])
gauge = aio.create_block(dashboard.key, gauge)

# Create a text stream
stream = Block(name="Stream Data",
               visual_type = 'stream',
               properties = {
                   "fontSize": "12",
                   "fontColor": "#63de00",
                   "showGroupName": "no"},
               block_feeds = [{
                   "group_id": group.id,
                   "feed_id":  feed.id
               }])
stream = aio.create_block(dashboard.key, stream)

# Update the large layout to:
# |----------------|
# |   Line Chart   |
# |----------------|
# | Gauge | Stream |
# |----------------|
layout = Layout(lg = [
                   {'x': 0, 'y': 0, 'w': 16, 'h': 4, 'i': str(linechart.id)},
                   {'x': 0, 'y': 4, 'w':  8, 'h': 4, 'i': str(gauge.id)},
                   {'x': 8, 'y': 4, 'w':  8, 'h': 4, 'i': str(stream.id)}])
aio.update_layout(dashboard.key, layout)

print("Dashboard created at: " +
      "https://io.adafruit.com/{0}/dashboards/{1}".format(ADAFRUIT_IO_USERNAME,
                                                          dashboard.key))
# Now send some data
value = 0
while True:
    value = (value + randrange(0, 10)) % 100
    print('sending data: ', value)
    aio.send_data(feed.key, value)
    sleep(3)



"""
`rgb_led.py`
=======================================================================
Control a RGB LED using
Adafruit IO and Python

Tutorial Link: https://learn.adafruit.com/adafruit-io-basics-color

Adafruit invests time and resources providing this open source code.
Please support Adafruit and open source hardware by purchasing
products from Adafruit!

Author(s): Brent Rubell for Adafruit Industries
Copyright (c) 2018 Adafruit Industries
Licensed under the MIT license.
All text above must be included in any redistribution.

Dependencies:
    - Adafruit_Blinka
        (https://github.com/adafruit/Adafruit_Blinka)
    - Adafruit_CircuitPython_PCA9685
        (https://github.com/adafruit/Adafruit_CircuitPython_PCA9685)
"""
# import system libraries
import time

# import Adafruit Blinka
from board import SCL, SDA
from busio import I2C

# import the PCA9685 module.
from adafruit_pca9685 import PCA9685

# import Adafruit IO REST client
from Adafruit_IO import Client, Feed, RequestError

# PWM Pins
RED_PIN = 6
GREEN_PIN = 5
BLUE_PIN = 4

# Set to your Adafruit IO key.
# Remember, your key is a secret,
# so make sure not to publish it when you publish this code!
ADAFRUIT_IO_KEY = 'YOUR_AIO_KEY'

# Set to your Adafruit IO username.
# (go to https://accounts.adafruit.com to find your username)
ADAFRUIT_IO_USERNAME = 'YOUR_AIO_USERNAME'

# Create an instance of the REST client.
aio = Client(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

try: # if we have a 'color' feed
    color = aio.feeds('color')
except RequestError: # create an `color` feed
    feed = Feed(name='color')
    color = aio.create_feed(feed)

# Create the I2C bus interface.
i2c_bus = I2C(SCL, SDA)

# Create a simple PCA9685 class instance.
pca = PCA9685(i2c_bus)
pca.frequency = 60
prev_color = '#000000'

def map_range(x, in_min, in_max, out_min, out_max):
    """re-maps a number from one range to another."""
    mapped = (x-in_min) * (out_max - out_min) / (in_max-in_min) + out_min
    if out_min <= out_max:
        return max(min(mapped, out_max), out_min)
    return min(max(mapped, out_max), out_min)

while True:
    # grab the `color` feed
    color_val = aio.receive(color.key)
    if color_val != prev_color:
        # print rgb values and hex value
        print('Received Color: ')
        red = aio.to_red(color_val.value)
        print('\t - R: ', red)
        green = aio.to_green(color_val.value)
        print('\t - G: ', green)
        blue = aio.to_blue(color_val.value)
        print('\t - B: ', blue)
        print('\t - HEX: ', color_val.value)
        # map color values (0-255) to  16-bit values for the pca
        red = map_range(int(red), 0, 255, 0, 65535)
        green = map_range(int(green), 0, 255, 0, 65535)
        blue = map_range(int(blue), 0, 255, 0, 65535)
        # invert RGB values for common anode LEDs.
        pca.channels[RED_PIN].duty_cycle = 65535 - int(red)
        pca.channels[GREEN_PIN].duty_cycle = 65535 - int(green)
        pca.channels[BLUE_PIN].duty_cycle = 65535 - int(blue)
    prev_color = color_val
    # let's wait a bit so we don't flood adafruit io's servers...
    time.sleep(1)



# Simple example of sending and receiving values from Adafruit IO with the REST
# API client.
# Author: Tony Dicola, Justin Cooper, Brent Rubell

# Import Adafruit IO REST client.
from Adafruit_IO import Client, Feed
import json

# Set to your Adafruit IO key.
# Remember, your key is a secret,
# so make sure not to publish it when you publish this code!
ADAFRUIT_IO_KEY = 'YOUR_AIO_KEY'

# Set to your Adafruit IO username.
# (go to https://accounts.adafruit.com to find your username)
ADAFRUIT_IO_USERNAME = 'YOUR_AIO_USERNAME'

# Create an instance of the REST client.
aio = Client(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

# List all of your feeds
print("Obtaining user's feeds...")
feeds = aio.feeds()
print('Feeds: ', feeds)

# Create a new feed
print("Creating new feed...")
feed = Feed(name="PythonFeed")
response = aio.create_feed(feed)
print("New feed: ", response)

# Delete a feed
aio.delete_feed(response.key)

# Create feed in a group
feed = Feed(name="PythonGroupFeed")
group_key = "example"
print("Creating feed in group %s"%group_key)
response = aio.create_feed(feed, group_key)
print("New feed: ", response)

# Delete a feed within a group
print("Deleting feed within group %s"%group_key)
aio.delete_feed(response.key)


"""
'analog_in.py'
==================================
Example of sending analog sensor
values to an Adafruit IO feed.

Author(s): Brent Rubell

Dependencies:
    - Adafruit_Blinka
        (https://github.com/adafruit/Adafruit_Blinka)
    - Adafruit_CircuitPython_MCP3xxx
        (https://github.com/adafruit/Adafruit_CircuitPython_MCP3xxx)
"""
# Import standard python modules
import time

# import Adafruit Blinka
import board
import digitalio
import busio

# import Adafruit IO REST client
from Adafruit_IO import Client, Feed, RequestError

# import Adafruit CircuitPython MCP3xxx library
from adafruit_mcp3xxx.mcp3008 import MCP3008
from adafruit_mcp3xxx.analog_in import AnalogIn

# Set to your Adafruit IO key.
# Remember, your key is a secret,
# so make sure not to publish it when you publish this code!
ADAFRUIT_IO_KEY = 'YOUR_AIO_KEY'

# Set to your Adafruit IO username.
# (go to https://accounts.adafruit.com to find your username)
ADAFRUIT_IO_USERNAME = 'YOUR_AIO_USERNAME'
# Create an instance of the REST client
aio = Client(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

try: # if we have a 'analog' feed
    analog = aio.feeds('analog')
except RequestError: # create a analog feed
    feed = Feed(name='analog')
    analog = aio.create_feed(feed)

# Create an instance of the `busio.spi` class
spi = busio.SPI(board.SCLK, board.MOSI, board.MISO)

# create the cs (chip select)
cs = digitalio.DigitalInOut(board.D12)

# create a mcp3008 object
mcp = MCP3008(spi, cs)

# create an an adc (single-ended) on pin 0
chan = AnalogIn(mcp, MCP3008.pin_0)

while True:
    sensor_data = chan.value

    print('Analog Data -> ', sensor_data)
    aio.send(analog.key, sensor_data)

    # avoid timeout from adafruit io
    time.sleep(0.5)



"""
'random_data.py'
================================================
Example for accessing the Adafruit IO Random
Data Service.

Author(s): Brent Rubell for Adafruit Industries
"""
# Import JSON for forecast parsing
import json
# Import Adafruit IO REST client.
from Adafruit_IO import Client, Feed, RequestError

# Set to your Adafruit IO key.
ADAFRUIT_IO_USERNAME = 'YOUR_IO_USERNAME'
ADAFRUIT_IO_KEY = 'YOUR_IO_KEY'

# Create an instance of the REST client.
aio = Client(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

generator_id = 1461

# Get the specified randomizer record with its current value and related details.
random_data = aio.receive_random(generator_id)
# Parse the API response
data = json.dumps(random_data)
data = json.loads(data)
print('Random Data: {0}'.format(data['value']))



# Copyright (c) 2018 Adafruit Industries
# Authors: Justin Cooper & Tony DiCola

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import time
from time import struct_time
import json
import platform
import pkg_resources
import re
from urllib.parse import urlparse
from urllib.parse import parse_qs
# import logging

import requests

from .errors import RequestError, ThrottlingError
from .model import Data, Feed, Group, Dashboard, Block, Layout

DEFAULT_PAGE_LIMIT = 100

# set outgoing version, pulled from setup.py
version = pkg_resources.require("Adafruit_IO")[0].version
default_headers = {
    'User-Agent': 'AdafruitIO-Python/{0} ({1}, {2} {3})'.format(version,
                                                                platform.platform(),
                                                                platform.python_implementation(),
                                                                platform.python_version())
}

class Client(object):
    """Client instance for interacting with the Adafruit IO service using its
    REST API.  Use this client class to send, receive, and enumerate feed data.
    """

    def __init__(self, username, key, proxies=None, base_url='https://io.adafruit.com'):
        """Create an instance of the Adafruit IO REST API client.  Key must be
        provided and set to your Adafruit IO access key value.  Optionaly
        provide a proxies dict in the format used by the requests library,
        and a base_url to point at a different Adafruit IO service
        (the default is the production Adafruit IO service over SSL).
        """
        self.username = username
        self.key = key
        self.proxies = proxies
        # self.logger = logging.basicConfig(level=logging.DEBUG,
        #                                   format='%(asctime)s - %(levelname)s - %(message)s')

        # Save URL without trailing slash as it will be added later when
        # constructing the path.
        self.base_url = base_url.rstrip('/')

        # Store the last response of a get or post
        self._last_response = None

    @staticmethod
    def to_red(data):
        """Hex color feed to red channel.
        
        :param int data: Color value, in hexadecimal.
        """
        return ((int(data[1], 16))*16) + int(data[2], 16)

    @staticmethod
    def to_green(data):
        """Hex color feed to green channel.

        :param int data: Color value, in hexadecimal.
        """
        return (int(data[3], 16) * 16) + int(data[4], 16)

    @staticmethod
    def to_blue(data):
        """Hex color feed to blue channel.

        :param int data: Color value, in hexadecimal.
        """
        return (int(data[5], 16) * 16) + int(data[6], 16)

    @staticmethod
    def _headers(given):
        headers = default_headers.copy()
        headers.update(given)
        return headers

    @staticmethod
    def _create_payload(value, metadata):
        if metadata is not None:
            payload = Data(value=value, lat=metadata['lat'], lon=metadata['lon'],
                           ele=metadata['ele'], created_at=metadata['created_at'])
            return payload
        return Data(value=value)

    @staticmethod
    def _handle_error(response):
        # Throttling Error
        if response.status_code == 429:
            raise ThrottlingError()
        # Resource on AdafruitIO not Found Error
        elif response.status_code == 400:
            raise RequestError(response)
        # Handle all other errors (400 & 500 level HTTP responses)
        elif response.status_code >= 400:
            raise RequestError(response)
        # Else do nothing if there was no error.

    def _compose_url(self, path):
        return '{0}/api/{1}/{2}/{3}'.format(self.base_url, 'v2', self.username, path)

    def _get(self, path, params=None):
        response = requests.get(self._compose_url(path),
                                headers=self._headers({'X-AIO-Key': self.key}),
                                proxies=self.proxies,
                                params=params)
        self._last_response = response
        self._handle_error(response)
        return response.json()

    def _post(self, path, data):
        response = requests.post(self._compose_url(path),
                                 headers=self._headers({'X-AIO-Key': self.key,
                                                        'Content-Type': 'application/json'}),
                                 proxies=self.proxies,
                                 data=json.dumps(data))
        self._last_response = response
        self._handle_error(response)
        return response.json()

    def _delete(self, path):
        response = requests.delete(self._compose_url(path),
                                   headers=self._headers({'X-AIO-Key': self.key,
                                                          'Content-Type': 'application/json'}),
                                   proxies=self.proxies)
        self._last_response = response
        self._handle_error(response)

    # Data functionality.
    def send_data(self, feed, value, metadata=None, precision=None):
        """Helper function to simplify adding a value to a feed.  Will append the
        specified value to the feed identified by either name, key, or ID.
        Returns a Data instance with details about the newly appended row of data.
        Note that send_data now operates the same as append.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        :param string value: Value to send.
        :param dict metadata: Optional metadata associated with the value.
        :param int precision: Optional amount of precision points to send.
        """
        if precision:
            try:
                value = round(value, precision)
            except NotImplementedError:
                raise NotImplementedError("Using the precision kwarg requires a float value")
        payload = self._create_payload(value, metadata)
        return self.create_data(feed, payload)

    send = send_data

    def send_batch_data(self, feed, data_list):
        """Create a new row of data in the specified feed.  Feed can be a feed
        ID, feed key, or feed name.  Data must be an instance of the Data class
        with at least a value property set on it.  Returns a Data instance with
        details about the newly appended row of data.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        :param Data data_list: Multiple data values.
        """
        path = "feeds/{0}/data/batch".format(feed)
        data_dict = type(data_list)((data._asdict() for data in data_list))
        self._post(path, {"data": data_dict})

    def append(self, feed, value):
        """Helper function to simplify adding a value to a feed.  Will append the
        specified value to the feed identified by either name, key, or ID.
        Returns a Data instance with details about the newly appended row of data.
        Note that unlike send the feed should exist before calling append.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        :param string value: Value to append to feed.
        """
        return self.create_data(feed, Data(value=value))

    def receive_time(self, timezone=None):
        """Returns a struct_time from the Adafruit IO Server based on requested
        timezone, or automatically based on the device's IP address.
        https://docs.python.org/3.7/library/time.html#time.struct_time

        :param string timezone: Optional timezone to return the time in.
        See https://en.wikipedia.org/wiki/List_of_tz_database_time_zones#List
        """
        path = 'integrations/time/struct.json'
        if timezone:
            path += f'?tz={timezone}'
        return self._parse_time_struct(self._get(path))

    @staticmethod
    def _parse_time_struct(time_dict: dict) -> time.struct_time:
        """Parse the time data returned by the server and return a time_struct

        Corrects for the weekday returned by the server in Sunday=0 format
        (Python expects Monday=0)
        """
        wday = (time_dict['wday'] - 1) % 7
        return struct_time((time_dict['year'], time_dict['mon'], time_dict['mday'],
                            time_dict['hour'], time_dict['min'], time_dict['sec'],
                            wday, time_dict['yday'], time_dict['isdst']))

    def receive_weather(self, weather_id=None):
        """Adafruit IO Weather Service, Powered by Dark Sky

        :param int id: optional ID for retrieving a specified weather record.
        """
        if weather_id:
            weather_path = "integrations/weather/{0}".format(weather_id)
        else:
            weather_path = "integrations/weather"
        return self._get(weather_path)

    def receive_random(self, randomizer_id=None):
        """Access to Adafruit IO's Random Data
        service.

        :param int randomizer_id: optional ID for retrieving a specified randomizer.
        """
        if randomizer_id:
            random_path = "integrations/words/{0}".format(randomizer_id)
        else:
            random_path = "integrations/words"
        return self._get(random_path)

    def receive(self, feed):
        """Retrieve the most recent value for the specified feed. Returns a Data
        instance whose value property holds the retrieved value.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        """
        path = "feeds/{0}/data/last".format(feed)
        return Data.from_dict(self._get(path))

    def receive_next(self, feed):
        """Retrieve the next unread value from the specified feed. Returns a Data
        instance whose value property holds the retrieved value.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        """
        path = "feeds/{0}/data/next".format(feed)
        return Data.from_dict(self._get(path))

    def receive_previous(self, feed):
        """Retrieve the previous unread value from the specified feed. Returns a
        Data instance whose value property holds the retrieved value.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        """
        path = "feeds/{0}/data/previous".format(feed)
        return Data.from_dict(self._get(path))

    def data(self, feed, data_id=None, max_results=DEFAULT_PAGE_LIMIT):
        """Retrieve data from a feed. If data_id is not specified then all the data
        for the feed will be returned in an array.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        :param string data_id: ID of the piece of data to delete.
        :param int max_results: The maximum number of results to return. To
            return all data, set to None.
        """
        if max_results is None:
            res = self._get(f'feeds/{feed}/details')
            max_results = res['details']['data']['count']
        if data_id:
            path = "feeds/{0}/data/{1}".format(feed, data_id)
            return Data.from_dict(self._get(path))

        params = {'limit': max_results} if max_results else None
        data = []
        path = "feeds/{0}/data".format(feed)
        while len(data) < max_results:
            data.extend(list(map(Data.from_dict, self._get(path,
                                                           params=params))))
            nlink = self.get_next_link()
            if not nlink:
                break
            # Parse the link for the query parameters
            params = parse_qs(urlparse(nlink).query)
            if max_results:
                params['limit'] = max_results - len(data)
        return data

    def get_next_link(self):
        """Parse the `next` page URL in the pagination Link header.

        This is necessary because of a bug in the API's implementation of the
        link header. If that bug is fixed, the link would be accesible by
        response.links['next']['url'] and this method would be broken.

        :return: The url for the next page of data
        :rtype: str
        """
        if not self._last_response:
            return
        link_header = self._last_response.headers['link']
        res = re.search('rel="next", <(.+?)>', link_header)
        if not res:
            return
        return res.groups()[0]

    def create_data(self, feed, data):
        """Create a new row of data in the specified feed.
        Returns a Data instance with details about the newly
        appended row of data.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        :param Data data: Instance of the Data class. Must have a value property set.
        """
        path = "feeds/{0}/data".format(feed)
        return Data.from_dict(self._post(path, data._asdict()))

    def delete(self, feed, data_id):
        """Delete data from a feed.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        :param string data_id: ID of the piece of data to delete.
        """
        path = "feeds/{0}/data/{1}".format(feed, data_id)
        self._delete(path)

    # feed functionality.
    def feeds(self, feed=None):
        """Retrieve a list of all feeds, or the specified feed.  If feed is not
        specified a list of all feeds will be returned.

        :param string feed: Name/Key/ID of Adafruit IO feed, defaults to None.
        """
        if feed is None:
            path = "feeds"
            return list(map(Feed.from_dict, self._get(path)))
        path = "feeds/{0}".format(feed)
        return Feed.from_dict(self._get(path))

    def create_feed(self, feed, group_key=None):
        """Create the specified feed.

        :param string feed: Key of Adafruit IO feed.
        :param group_key group: Group to place new feed in.
        """
        f = feed._asdict()
        del f['id']  # Don't pass id on create call
        path = "feeds/"
        if group_key is not None: # create feed in a group
            path="/groups/%s/feeds"%group_key
            return Feed.from_dict(self._post(path, {"feed": f}))
        return Feed.from_dict(self._post(path, {"feed": f}))

    def delete_feed(self, feed):
        """Delete the specified feed.

        :param string feed: Name/Key/ID of Adafruit IO feed.
        """
        path = "feeds/{0}".format(feed)
        self._delete(path)

    # Group functionality.
    def groups(self, group=None):
        """Retrieve a list of all groups, or the specified group.

        :param string group: Name/Key/ID of Adafruit IO Group. Defaults to None.
        """
        if group is None:
            path = "groups/"
            return list(map(Group.from_dict, self._get(path)))
        path = "groups/{0}".format(group)
        return Group.from_dict(self._get(path))

    def create_group(self, group):
        """Create the specified group.

        :param string group: Name/Key/ID of Adafruit IO Group.
        """
        path = "groups/"
        return Group.from_dict(self._post(path, group._asdict()))

    def delete_group(self, group):
        """Delete the specified group.

        :param string group: Name/Key/ID of Adafruit IO Group.
        """
        path = "groups/{0}".format(group)
        self._delete(path)

    # Dashboard functionality.
    def dashboards(self, dashboard=None):
        """Retrieve a list of all dashboards, or the specified dashboard.

        :param string dashboard: Key of Adafruit IO Dashboard. Defaults to None.
        """
        if dashboard is None:
            path = "dashboards/"
            return list(map(Dashboard.from_dict, self._get(path)))
        path = "dashboards/{0}".format(dashboard)
        return Dashboard.from_dict(self._get(path))

    def create_dashboard(self, dashboard):
        """Create the specified dashboard.

        :param Dashboard dashboard: Dashboard object to create
        """
        path = "dashboards/"
        return Dashboard.from_dict(self._post(path, dashboard._asdict()))

    def delete_dashboard(self, dashboard):
        """Delete the specified dashboard.

        :param string dashboard: Key of Adafruit IO Dashboard.
        """
        path = "dashboards/{0}".format(dashboard)
        self._delete(path)

    # Block functionality.
    def blocks(self, dashboard, block=None):
        """Retrieve a list of all blocks from a dashboard, or the specified block.

        :param string dashboard: Key of Adafruit IO Dashboard.
        :param string block: id of Adafruit IO Block. Defaults to None.
        """
        if block is None:
            path = "dashboards/{0}/blocks".format(dashboard)
            return list(map(Block.from_dict, self._get(path)))
        path = "dashboards/{0}/blocks/{1}".format(dashboard, block)
        return Block.from_dict(self._get(path))

    def create_block(self, dashboard, block):
        """Create the specified block under the specified dashboard.

        :param string dashboard: Key of Adafruit IO Dashboard.
        :param Block block: Block object to create under dashboard
        """
        path = "dashboards/{0}/blocks".format(dashboard)
        return Block.from_dict(self._post(path, block._asdict()))

    def delete_block(self, dashboard, block):
        """Delete the specified block.

        :param string dashboard: Key of Adafruit IO Dashboard.
        :param string block: id of Adafruit IO Block.
        """
        path = "dashboards/{0}/blocks/{1}".format(dashboard, block)
        self._delete(path)

    # Layout functionality.
    def layouts(self, dashboard):
        """Retrieve the layouts array from a dashboard

        :param string dashboard: key of Adafruit IO Dashboard.
        """
        path = "dashboards/{0}".format(dashboard)
        dashboard = self._get(path)
        return Layout.from_dict(dashboard['layouts'])

    def update_layout(self, dashboard, layout):
        """Update the layout of the specified dashboard.

        :param string dashboard: Key of Adafruit IO Dashboard.
        :param Layout layout: Layout object to update under dashboard
        """
        path = "dashboards/{0}/update_layouts".format(dashboard)
        return Layout.from_dict(self._post(path, {'layouts': layout._asdict()}))



"""
`rgb_led.py`
=======================================================================
Control a NeoPixel RGB LED using Adafruit IO and Python

Tutorial Link: https://learn.adafruit.com/adafruit-io-basics-color

Adafruit invests time and resources providing this open source code.
Please support Adafruit and open source hardware by purchasing
products from Adafruit!

Author(s): Brent Rubell for Adafruit Industries
Copyright (c) 2023 Adafruit Industries
Licensed under the MIT license.
All text above must be included in any redistribution.

Dependencies:
    - Adafruit_Blinka
        (https://github.com/adafruit/Adafruit_Blinka)
    - Adafruit_CircuitPython_NeoPixel
        (https://github.com/adafruit/Adafruit_CircuitPython_NeoPixel)
"""
import time
import board
import neopixel
from Adafruit_IO import Client, Feed, RequestError

# Choose an open pin connected to the Data In of the NeoPixel strip, i.e. board.D18
# NeoPixels must be connected to D10, D12, D18 or D21 to work.
pixel_pin = board.D18

# The number of NeoPixels
num_pixels = 1

# The order of the pixel colors - RGB or GRB. Some NeoPixels have red and green reversed!
ORDER = neopixel.GRB

pixels = neopixel.NeoPixel(
    pixel_pin, num_pixels, brightness=0.2, auto_write=False, pixel_order=ORDER
)

# Set to your Adafruit IO key.
# Remember, your key is a secret,
# so make sure not to publish it when you publish this code!
ADAFRUIT_IO_KEY = 'YOUR_AIO_KEY'

# Set to your Adafruit IO username.
# (go to https://accounts.adafruit.com to find your username)
ADAFRUIT_IO_USERNAME = 'YOUR_AIO_USERNAME'

# Create an instance of the REST client.
aio = Client(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

try: # if we have a 'color' feed
    color = aio.feeds('color')
except RequestError: # create an `color` feed
    feed = Feed(name='color')
    color = aio.create_feed(feed)

while True:
    # get the value of the Adafruit IO `color` feed
    color_val = aio.receive(color.key)
    # Print hex value
    print('Received Color HEX: ', color_val)
    pixels.fill(color_val.value)
    pixels.show()

    # let's sleep/wait so we don't flood adafruit io's servers with requests
    time.sleep(3)



"""
'temp_humidity.py'
==================================
Example of sending temperature and humidity data to Adafruit IO

Author(s): Brent Rubell

Tutorial Link: Tutorial Link: https://learn.adafruit.com/adafruit-io-basics-temperature-and-humidity

Dependencies:
    - Adafruit IO Python Client
        (https://github.com/adafruit/io-client-python)
    - Adafruit_CircuitPython_AHTx0
        (https://github.com/adafruit/Adafruit_CircuitPython_AHTx0)
"""

# import standard python modules.
import time

# import adafruit-blinka modules
import board

# import Adafruit IO REST client.
from Adafruit_IO import Client, Feed, RequestError

# Import AHTx0 library
import adafruit_ahtx0

# Set true to send tempertaure data in degrees fahrenheit ('f')?
USE_DEGREES_F = False

# Time between sensor reads, in seconds
READ_TIMEOUT = 60

# Set to your Adafruit IO key.
# Remember, your key is a secret,
# so make sure not to publish it when you publish this code!
ADAFRUIT_IO_KEY = 'YOUR_AIO_KEY'

# Set to your Adafruit IO username.
# (go to https://accounts.adafruit.com to find your username).
ADAFRUIT_IO_USERNAME = 'YOUR_AIO_USERNAME'

# Create an instance of the REST client.
aio = Client(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

# Assign a temperature feed, if one exists already
try:
    temperature_feed = aio.feeds('temperature')
except RequestError: # Doesn't exist, create a new feed
    feed_temp = Feed(name="temperature")
    temperature_feed = aio.create_feed(feed_temp)

# Assign a humidity feed, if one exists already
try:
    humidity_feed = aio.feeds('humidity')
except RequestError: # Doesn't exist, create a new feed
    feed_humid = Feed(name="humidity")
    humidity_feed = aio.create_feed(feed_humid)

# Initialize the board's default I2C bus
i2c = board.I2C()  # uses board.SCL and board.SDA
# Initialize AHT20 using the default address (0x38) and the board's default i2c bus
sensor = adafruit_ahtx0.AHTx0(i2c)

while True:
    temperature = sensor.temperature
    humidity = sensor.relative_humidity
    if USE_DEGREES_F:
        temperature = temperature * 9.0 / 5.0 + 32.0
        print('Temp={0:0.1f}*F'.format(temperature))
    else:
        print('Temp={0:0.1f}*C'.format(temperature))
    print('Humidity={1:0.1f}%'.format(humidity))
    # Format sensor data as string for sending to Adafruit IO
    temperature = '%.2f'%(temperature)
    humidity = '%.2f'%(humidity)
    # Send humidity and temperature data to Adafruit IO
    aio.send(temperature_feed.key, str(temperature))
    aio.send(humidity_feed.key, str(humidity))

    # Timeout to avoid flooding Adafruit IO
    time.sleep(READ_TIMEOUT)



"""
`analog_output.py`
===============================================================================
PWM a LED from Adafruit IO!

Tutorial Link: https://learn.adafruit.com/adafruit-io-basics-analog-output

Adafruit invests time and resources providing this open source code.
Please support Adafruit and open source hardware by purchasing
products from Adafruit!

Author(s): Brent Rubell for Adafruit Industries
Copyright (c) 2018 Adafruit Industries
Licensed under the MIT license.
All text above must be included in any redistribution.

Dependencies:
    - Adafruit_Blinka
        (https://github.com/adafruit/Adafruit_Blinka)
    - Adafruit_CircuitPython_PCA9685
        (https://github.com/adafruit/Adafruit_CircuitPython_PCA9685)
"""

# import system libraries
import time

# import Adafruit Blinka
from board import SCL, SDA
from busio import I2C

# import the PCA9685 module.
from adafruit_pca9685 import PCA9685

# import Adafruit IO REST client
from Adafruit_IO import Client, Feed, RequestError

# Set to your Adafruit IO key.
# Remember, your key is a secret,
# so make sure not to publish it when you publish this code!
ADAFRUIT_IO_KEY = 'YOUR_IO_KEY'

# Set to your Adafruit IO username.
# (go to https://accounts.adafruit.com to find your username)
ADAFRUIT_IO_USERNAME = 'YOUR_IO_USERNAME'

# Create the I2C bus interface.
i2c_bus = I2C(SCL, SDA)

# Create a simple PCA9685 class instance.
pca = PCA9685(i2c_bus)
PCA_CHANNEL = 4

# Set the PWM frequency to 60hz.
pca.frequency = 60
prev_read = 0

# Create an instance of the REST client.
aio = Client(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

try: # if we have a 'analog' feed
    analog = aio.feeds('analog')
except RequestError: # create an `analog` feed
    feed = Feed(name='analog')
    analog = aio.create_feed(feed)


def map_range(x, in_min, in_max, out_min, out_max):
    """re-maps a number from one range to another."""
    mapped = (x-in_min) * (out_max - out_min) / (in_max-in_min) + out_min
    if out_min <= out_max:
        return max(min(mapped, out_max), out_min)
    return min(max(mapped, out_max), out_min)

while True:
    # grab the `analog` feed value
    analog_read = aio.receive(analog.key)
    if analog_read.value != prev_read:
        print('received <- ', analog_read.value)
        # map the analog value from 0 - 1023 to 0 - 65534
        analog_value = map_range(int(analog_read.value), 0, 1024, 0, 65534)
        # set the LED to the mapped feed value
        pca.channels[PCA_CHANNEL].duty_cycle = int(analog_value)
    prev_read = analog_read.value
    # timeout so we don't flood IO with requests
    time.sleep(0.5)



"""
`mqtt_shared_feeds.py`
---------------------------------------------------------
Example of reading and writing to a shared  Adafruit IO Feed.

learn.adafruit.com/adafruit-io-basics-feeds/sharing-a-feed

Author: Brent Rubell for Adafruit Industries 2018
"""

# Import standard python modules.
import sys
import time
import random

# Import Adafruit IO MQTT client.
from Adafruit_IO import MQTTClient

# Set to your Adafruit IO key.
# Remember, your key is a secret,
# so make sure not to publish it when you publish this code!
ADAFRUIT_IO_KEY = 'YOUR_AIO_KEY'

# Set to your Adafruit IO username.
# (go to https://accounts.adafruit.com to find your username)
ADAFRUIT_IO_USERNAME = 'YOUR_AIO_USERNAME'

# Shared IO Feed
# Make sure you have read AND write access to this feed to publish.
IO_FEED = 'SHARED_AIO_FEED_NAME'

# IO Feed Owner's username
IO_FEED_USERNAME = 'SHARED_AIO_FEED_USERNAME'


# Define callback functions which will be called when certain events happen.
def connected(client):
    """Connected function will be called when the client connects.
    """
    client.subscribe(IO_FEED, IO_FEED_USERNAME)

def disconnected(client):
    """Disconnected function will be called when the client disconnects.
    """
    print('Disconnected from Adafruit IO!')
    sys.exit(1)

def message(client, feed_id, payload):
    """Message function will be called when a subscribed feed has a new value.
    The feed_id parameter identifies the feed, and the payload parameter has
    the new value.
    """
    print('Feed {0} received new value: {1}'.format(feed_id, payload))


# Create an MQTT client instance.
client = MQTTClient(ADAFRUIT_IO_USERNAME, ADAFRUIT_IO_KEY)

# Setup the callback functions defined above.
client.on_connect       =   connected
client.on_disconnect    =   disconnected
client.on_message       =   message

# Connect to the Adafruit IO server.
client.connect()

client.loop_background()
print('Publishing a new message every 10 seconds (press Ctrl-C to quit)...')

while True:
    value = random.randint(0, 100)
    print('Publishing {0} to {1}.'.format(value, IO_FEED))
    client.publish(IO_FEED, value, feed_user=IO_FEED_USERNAME)
    time.sleep(10)
